/***************************************************************************
    NWNX System - System functions plugin
    Copyright (C) 2006 Ingmar Stieger (Papillon, papillon@nwnx.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ***************************************************************************/

/*

	---------------------------------------------------------------------------------------
	Date		Name		Reason
	---------------------------------------------------------------------------------------
    
    18 Jan 07   Edward      Initial release.
    05 Aug 07   Edward      Refactored to NWNX4 1.08.
    25 Aug 07   Edward      Support for multiple instances of NWN2 server.
    25 Aug 07   Edward      Added more useful error messages.
    08 Dec 07   Edward      Added a more robust server reset function.
    12 Nov 08   Edward      Recompiled for NWN2 patch 1.13 and NWNX 1.09.
    06 Apr 09   Edward      Added NWNX4 1.09 folder support.
                            CPU Core configuration is now found relative to the plugin's folder.
    07 Apr 09   Edward      Added CPU usage function.

	---------------------------------------------------------------------------------------

*/


// Defines.
#define WIN32_LEAN_AND_MEAN


// Includes.
#include "system.h"


/***************************************************************************
    NWNX and DLL specific functions
***************************************************************************/

System * plugin;

DLLEXPORT Plugin * GetPluginPointerV2( ){
	return plugin;
}


BOOL APIENTRY DllMain( HMODULE hModule, DWORD dwReason, LPVOID lpReserved ){

	switch( dwReason ){

		case DLL_PROCESS_ATTACH:{
			plugin = new System( );
			TCHAR szPath[ MAX_PATH ];
			GetModuleFileName( hModule, szPath, MAX_PATH );
			plugin->SetPluginFullPath( szPath );
			break;
								}
		case DLL_PROCESS_DETACH:{
			delete plugin;
			break;
								}
		default:
			break;

	}

	return( TRUE );

}


/***************************************************************************
    Implementation of System Plugin
***************************************************************************/


System :: System( ){

	header		=	"NWNX System Plugin V.";
	header	   +=	SYSTEM_VERSION;
	header	   +=	"\n(c) 2006 by Ingmar Stieger (Papillon)\nvisit us at http://www.nwnx.org\n";

	description	=	"This plugin provides system functions.";

	subClass	=	"SYSTEM";

	version		=	SYSTEM_VERSION;

	return;

}


System :: ~System( ){

	// Cleanup.
	CloseHandle( hProcess );

	wxLogMessage( wxT( "* NWNX4/System: Plugin unloaded." ) );

}


bool System :: Init( PCHAR ps_nwnxhome ){

	assert( GetPluginFileName( ) );

	// Log file.
	wxString logfile( ps_nwnxhome ); 
	logfile.append( wxT( "\\" ) );
	logfile.append( GetPluginFileName( ) );
	logfile.append( wxT( ".txt" ) );
	logger = new wxLogNWNX( logfile, wxString( header.c_str( ) ) );

	// Load the configuration.
    wxString pathToConfig( GetPluginFolder( ) );
    pathToConfig.append( wxT( "xp_system.ini" ) );

	config = new wxFileConfig(
			wxEmptyString,
			wxEmptyString,
            pathToConfig,
			wxEmptyString,
			wxCONFIG_USE_RELATIVE_PATH | wxCONFIG_USE_NO_ESCAPE_CHARACTERS );

	// Get the desired CPU Core configuration.
    wxLogMessage( "The file path to the Core configuration is: [" + pathToConfig + "]" );
	wxString sCpuID = config->Read( "cpuid" );
	// Convert it to an integer so that we can use it.
	sCpuID.ToULong( &uCPUID );

	// CPU Core config.
	if(			uCPUID == 0		)
		wxLogMessage( "* NWNX4/System: CPU configured to both cores." );

	else if(	uCPUID > 0 && uCPUID < 3	){
		// Open the NW2Server process for modification.
		hProcess = OpenProcess(
			PROCESS_ALL_ACCESS,
			FALSE,
			GetCurrentProcessId( ) );

		if( SetProcessAffinityMask( hProcess, uCPUID ) )
			wxLogMessage( "* NWNX4/System: CPU configured to %lu core.", uCPUID - 1 );
		else
			wxLogMessage( "* NWNX4/System: CPU core configuration failed. Please post this error to the NWNX forums: http://www.nwnx.org/phpBB2." );

	}

	else
		wxLogMessage( "* NWNX4/System: CPU core configuration failed. Please check your cpuid setting in the xp_system.ini file, which is in your root NWN2 folder." );

	wxLogMessage( wxT( "* NWNX4/System: Plugin initialized." ) );
	return( TRUE );

}


void System :: GetFunctionClass( TCHAR * fClass ){

	_tcsncpy_s( fClass, 128, wxT( "SYSTEM" ), 6 ); 

	return;

}


VOID System :: SetString( PCHAR psFunction, PCHAR psParam1, INT nParam2, PCHAR psValue ){

	// Log the NWNX4 string for debugging purposes.
	wxLogMessage(
		wxT( "* * NWNX4/System: Plugin SetString( psFunction = %s, psParam1 = %s, nParam2 = %d, psValue = %s )" ),
		psFunction,
		psParam1,
		nParam2,
		psValue );

	// Variables.
	wxString function( psFunction, wxConvUTF8 );
	wxString parameter1( psParam1, wxConvUTF8 );
	wxString parameter3( psValue, wxConvUTF8 );

	
	// Check for a function string.
	if(			function == wxT( "" ) ){
		wxLogMessage( wxT( "* NWNX4/System: Function not specified." ) );
		return;
	}

	// Execute the desired function.

	// ResetServer( .. ).
	if(			function == wxT( "RESET" ) )
		ResetServer( );

	return;

}


PCHAR System :: GetString( PCHAR psFunction, PCHAR psParam1, INT nParam2 ){

    // Log the NWNX4 string for debugging purposes.
	wxLogMessage(
		wxT( "* NWNX4/System: Plugin GetString( psFunction = %s, psParam1 = %s, nParam2 = %d ) " ),
		psFunction,
		psParam1,
		nParam2 );

	// Variables.
	wxString wxRequest( psFunction, wxConvUTF8 );
	wxString function( psFunction, wxConvUTF8 );
	wxString parameter1( psParam1, wxConvUTF8 );


	// Check for a function string.
	if(			function == wxT( "" ) ){
		wxLogMessage( wxT( "* NWNX4/System: Function not specified." ) );
		return( NULL );
	}




    // Function: GETCPU -> Gets the CPU usage.
    else if(    function == wxT( "GETCPU" ) )       GetCpuUsage( );




	// Name, Description, Subclass, Version query.
	else{

		// Process generic functions.
		wxString query = ProcessQueryFunction( function.c_str( ) );
		if( query != wxT( "" ) )
			sprintf_s( returnBuffer, MAX_BUFFER, "%s", query );
		else{
			wxLogMessage( wxT( "* NWNX4/System: Unknown function '%s' called." ), function );
			return( NULL );
		}
	
	}

	return( returnBuffer );

}


// This method will write text to NWN2's NWNXGetString(..) return value.
VOID System :: WriteNWN2String( PCHAR pszNWN2String ){

	// Copy the desired string into NWN2's one.
	nwnxcpy( returnBuffer, pszNWN2String );

	return;

}


// This method will close the NW2Server, NWNX4 can automatically restart it then.
VOID System :: ResetServer( ){

	// NW2Server handle invalid.
	if( hProcess == INVALID_HANDLE_VALUE )
		// Log.
		wxLogMessage( "* NWNX4/System: Can't find the NW2Server process. Please post this error to the NWNX forums: http://www.nwnx.org/phpBB2." );
	// NW2Server handle valid, reset the server by closing NW2Server down.
	else{
		// Log.
		wxLogMessage( "* NWNX4/System: NW2Server process found, restarting the server..." );
		// Open the NW2Server process for modification.
		hProcess = OpenProcess(
			PROCESS_ALL_ACCESS,
			FALSE,
			GetCurrentProcessId( ) );
		// Restart.
		TerminateProcess( hProcess, TRUE );
	}

	return;

}


// This method will gets the CPU usage as a percentage.
VOID System :: GetCpuUsage( ){

    // Variables.
    HQUERY      hQuery;
    HCOUNTER    hCounter;
    PDH_STATUS  pdhStatus;
    DWORD       dwType;


    // Open a performance data helper query.
    pdhStatus = PdhOpenQuery( NULL, NULL, &hQuery );
    if( pdhStatus != ERROR_SUCCESS ){
        wxLogMessage( "* NWNX4/System: Performance query error." );
        PdhCloseQuery( hQuery );
        return;
    }

    // Add the processor counter to the query.
    pdhStatus = PdhAddCounter( hQuery, "\\Processor(_Total)\\% Processor Time", 0, &hCounter );

    // Collect the query data.
    pdhStatus = PdhCollectQueryData( hQuery );
    if( pdhStatus != ERROR_SUCCESS ){
        PdhCloseQuery( hQuery );
    }

    // Wait 1 second.
    Sleep( 1000 );

    // Collect the query data again.
    pdhStatus = PdhCollectQueryData( hQuery );
    if( pdhStatus != ERROR_SUCCESS ){
        PdhCloseQuery( hQuery );
    }

    // Format the data value.
    pdhStatus = PdhGetFormattedCounterValue( hCounter, PDH_FMT_DOUBLE, &dwType, &pdhValue );

    if( pdhStatus != ERROR_SUCCESS ){
        wxLogMessage( "* NWNX4/System: Performance query formatting error." );
        PdhCloseQuery( hQuery );
        return;
    }

    // Return the percentage back to NWN2.
    CHAR szUsage[ 100 ]	= "";
    StringCbPrintf( szUsage, 100, "%.2f", pdhValue.doubleValue );
    WriteNWN2String( szUsage );

    // Cleanup.
    PdhCloseQuery( hQuery );
    return;

}